#ifdef NEVERDEF
/****************************************************************************
 ****************************************************************************
 *                                                                          *
 *   WARNING!                                                               *
 *                                                                          *
 *   I've discovered a bug in the code optimizer.  A definition like:       *
 *      #define TESTBITS (BIT1|BIT2)                                        *
 *   can produce incorrect code output from the optimizer when TESTBITS     *
 *   is used.                                                               *
 *                                                                          *
 *   Instead, use a definition like:                                        *
 *      #define TESTBITS 0x05  /* (BIT1|BIT2) */                            *
 *   Yes, that's not as "ideal", but it will save a WORLD of problems!      *
 *                                                                          *
 ****************************************************************************
 ****************************************************************************/
#endif

/*
 *  PIC emulator.  12C509A version.
 *  Register name definitions
 *
 *
 *  Copyright (c) 2001, 2002 by William Couture.
 *
 *  Released under the GNU GPL.  See http://www.gnu.org/licenses/gpl.txt
 *
 *  This program is part of PICEMU
 *
 *  PICEMU is free software; you can redistribute it and/or modify it under
 *  the terms of the GNU General Public License as published by the Free
 *  Software Foundatation; either version 2 of the License, or any later
 *  version.
 *
 *  PICEMU is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *
 *  Revision history
 *
 *  Programmer         Date      Comments
 *  -----------------  --------  ---------------------------------------------
 *  William Couture    9/xx/01   Original code
 *  William Couture    11/14/01  Code cleanup for 1st BETA release
 */

#define INDF    0x00
#define TMR0    0x01
#define PCL     0x02
#define STATUS  0x03
#define FSR     0x04
#define OSCCAL  0x05
#define GPIO    0x06

#define PORTA   GPIO          /* for more uniform code */
#define TRISA   (NUM_REGS)    /* hidden register */
#define OPTION  (NUM_REGS+1)  /* hidden register */

/*
 *  Now, do register bit name definitions as bit masks
 *
 *  I will not be doing all of them, as I don't use all of them.
 *
 *  Feel free to complete the list.
 */

   /* bits for STATUS (0x003) reg */
#define STATUS_C     0x01 /* bit mask */
#define STATUS_DC    0x02 /* bit mask */
#define STATUS_Z     0x04 /* bit mask */
#define STATUS_PD    0x08 /* bit mask (actually, !PD) */
#define STATUS_TO    0x10 /* bit mask (actually, !TO) */
#define STATUS_PA0   0x20 /* bit mask: page flag 0 (12-bit) */
#define STATUS_GPWUF 0x80 /* bit mask */

   /* bits for OPTION (optionreg) */
#define PSA    0x08   /* bit mask */
#define T0SE   0x10   /* bit mask */
#define T0CS   0x20   /* bit mask */
#define GPWU   0x80   /* bit mask */

   /* bits for GPIO (0x006) reg */
#define GP2    0x04   /* bit mask */
#define GP3    0x08   /* bit mask */
#define NOT_GP3 0xf7  /* bit mask */
