#ifdef NEVERDEF
/****************************************************************************
 ****************************************************************************
 *                                                                          *
 *   WARNING!                                                               *
 *                                                                          *
 *   I've discovered a bug in the code optimizer.  A definition like:       *
 *      #define TESTBITS (BIT1|BIT2)                                        *
 *   can produce incorrect code output from the optimizer when TESTBITS     *
 *   is used.                                                               *
 *                                                                          *
 *   Instead, use a definition like:                                        *
 *      #define TESTBITS 0x05  /* (BIT1|BIT2) */                            *
 *   Yes, that's not as "ideal", but it will save a WORLD of problems!      *
 *                                                                          *
 ****************************************************************************
 ****************************************************************************/
#endif

/*
 *  PIC emulator.  12C509A version.
 *  Register definitions
 *
 *  Copyright (c) 2001-2004 by William Couture.
 *
 *  Released under the GNU GPL.  See http://www.gnu.org/licenses/gpl.txt
 *
 *  This program is part of PICEMU
 *
 *  PICEMU is free software; you can redistribute it and/or modify it under
 *  the terms of the GNU General Public License as published by the Free
 *  Software Foundatation; either version 2 of the License, or any later
 *  version.
 *
 *  PICEMU is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *
 *  Revision history
 *
 *  Programmer         Date      Comments
 *  -----------------  --------  ---------------------------------------------
 *  William Couture    3/16/04   Split data from P12REGS.C file.
 */

#ifdef ALIGN_REGS_1
   static BYTE align_regs_1 = 0;
#endif  /* ALIGN_REGS_1 */
#ifdef ALIGN_REGS_2
   static WORD align_regs_2 = 0;
#endif  /* ALIGN_REGS_2 */

WORD xlate_regs[MAX_NUM_REGS] =
   {
      /* Bank 0 -- all regs map to themselves */
   INDF,   TMR0,   PCL,    STATUS, FSR,    OSCCAL, GPIO,   0x07,
   0x08,   0x09,   0x0a,   0x0b,   0x0c,   0x0d,   0x0e,   0x0f,

   0x10,   0x11,   0x12,   0x13,   0x14,   0x15,   0x16,   0x17,
   0x18,   0x19,   0x1a,   0x1b,   0x1c,   0x1d,   0x1e,   0x1f,

      /* Bank 1 -- first 16 regs map back to bank 0 */
   INDF,   TMR0,   PCL,    STATUS, FSR,    OSCCAL, GPIO,   0x07,
   0x08,   0x09,   0x0a,   0x0b,   0x0c,   0x0d,   0x0e,   0x0f,

   0x30,   0x31,   0x32,   0x33,   0x34,   0x35,   0x36,   0x37,
   0x38,   0x39,   0x3a,   0x3b,   0x3c,   0x3d,   0x3e,   0x3f
   };

/*
 *  Note:  There is no way to reach the functions which are mapped to another
 *         bank via the xlate_regs[] array.  They are there as placeholders.
 */
BYTE (*readregs[MAX_NUM_REGS])() =
   {
      /* BANK 0 */
   /* 00      01       02       03       04       05       06       07 */
   r_reg,   r_reg,   r_reg,   r_reg,   r_reg,   r_reg,   r_gpio,  r_reg,
   /* 08      09       0a       0b       0c       0d       0e       0f */
   r_reg,   r_reg,   r_reg,   r_reg,   r_reg,   r_reg,   r_reg,   r_reg,
   /* 10      11       12       13       14       15       16       17 */
   r_reg,   r_reg,   r_reg,   r_reg,   r_reg,   r_reg,   r_reg,   r_reg,
   /* 18      19       1a       1b       1c       1d       1e       1f */
   r_reg,   r_reg,   r_reg,   r_reg,   r_reg,   r_reg,   r_reg,   r_reg,
      /* BANK 1 */
   /* 20      21       22       23       24       25       26        27 */
   r_reg,   r_reg,   r_reg,   r_reg,   r_reg,   r_reg,   r_gpio,   r_reg,
   /* 28      29       2a       2b       2c       2d       2e       2f */
   r_reg,   r_reg,   r_reg,   r_reg,   r_reg,   r_reg,   r_reg,   r_reg,
   /* 30      31       32       33       34       35       36       37 */
   r_reg,   r_reg,   r_reg,   r_reg,   r_reg,   r_reg,   r_reg,   r_reg,
   /* 38      39       3a       3b       3c       3d       3e       3f */
   r_reg,   r_reg,   r_reg,   r_reg,   r_reg,   r_reg,   r_reg,   r_reg
   };

/*
 *  Note:  There is no way to reach the functions which are mapped to another
 *         bank via the xlate_regs[] array.  They are there as placeholders.
 */
void (*writeregs[MAX_NUM_REGS])() =
   {
      /* BANK 0 */
   /* 00      01       02       03       04       05       06       07 */
   w_none,  w_tmr0,  w_pcl,   w_status,w_fsr,   w_osccal,w_gpio,  w_reg,
   /* 08      09       0a       0b       0c       0d       0e       0f */
   w_reg,   w_reg,   w_reg,   w_reg,   w_reg,   w_reg,   w_reg,   w_reg,
   /* 10      11       12       13       14       15       16       17 */
   w_reg,   w_reg,   w_reg,   w_reg,   w_reg,   w_reg,   w_reg,   w_reg,
   /* 18      19       1a       1b       1c       1d       1e       1f */
   w_reg,   w_reg,   w_reg,   w_reg,   w_reg,   w_reg,   w_reg,   w_reg,
      /* BANK 1 */
   /* 20      21       22       23       24       25       26       27 */
   w_none,  w_tmr0,  w_pcl,   w_status,w_fsr,   w_osccal,w_gpio,  w_reg,
   /* 28      29       2a       2b       2c       2d       2e       2f */
   w_reg,   w_reg,   w_reg,   w_reg,   w_reg,   w_reg,   w_reg,   w_reg,
   /* 30      31       32       33       34       35       36       37 */
   w_reg,   w_reg,   w_reg,   w_reg,   w_reg,   w_reg,   w_reg,   w_reg,
   /* 38      39       3a       3b       3c       3d       3e       3f */
   w_reg,   w_reg,   w_reg,   w_reg,   w_reg,   w_reg,   w_reg,   w_reg
   };
