#ifdef NEVERDEF
/****************************************************************************
 ****************************************************************************
 *                                                                          *
 *   WARNING!                                                               *
 *                                                                          *
 *   I've discovered a bug in the code optimizer.  A definition like:       *
 *      #define TESTBITS (BIT1|BIT2)                                        *
 *   can produce incorrect code output from the optimizer when TESTBITS     *
 *   is used.                                                               *
 *                                                                          *
 *   Instead, use a definition like:                                        *
 *      #define TESTBITS 0x05  /* (BIT1|BIT2) */                            *
 *   Yes, that's not as "ideal", but it will save a WORLD of problems!      *
 *                                                                          *
 ****************************************************************************
 ****************************************************************************/
#endif

/*
 *  General include file for PICEMU project.
 *
 *
 *  Copyright (c) 2001-2004 by William Couture.
 *
 *  Released under the GNU GPL.  See http://www.gnu.org/licenses/gpl.txt
 *
 *  This program is part of PICEMU
 *
 *  PICEMU is free software; you can redistribute it and/or modify it under
 *  the terms of the GNU General Public License as published by the Free
 *  Software Foundatation; either version 2 of the License, or any later
 *  version.
 *
 *  PICEMU is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *
 *  Revision history
 *
 *  Programmer         Date      Comments
 *  -----------------  --------  ---------------------------------------------
 *  William Couture    9/xx/01   Original code
 *  William Couture    11/14/01  Code cleanup for 1st BETA release
 *  William Couture    5/23/02   Cleanup, it's getting too complex
 *  Matthias Melcher   9/6/02    Found bugs with #define's / #ifdef's
 *  (www.matthiasm.com)
 *  Mauro Giachero     1/3/03    Added some #define for keyboard handling
 *                               and READ_USES_TRIS
 *  William Couture    5/16/03   Update to add 12F675
 *  William Couture    3/15/04   Source code unification for 14-bit core
 *                               programs.
 */

typedef unsigned char BYTE;
typedef unsigned int WORD;
typedef long DWORD;  /* unsigned would be nice, but... */

#define TRUE 1
#define FALSE 0

#define MIN(a,b) (((a) < (b)) ? (a) : (b))
#define NULL (0)
#define MEMMOVE(s,d,l) _move(l,s,d)  /* better memonics */


#include <pictype.h>


/*
 *  Stuff for the display Windows
 */
#define VIDEO_NORMAL 0x0700  /* standard white on black */
#define VIDEO_HIGHLIGHT 0x0e00  /* yellow on black */
#define WINDOW_COMMAND 0   /* handled differently, but I want some of the data */
#define WINDOW_INSTR   1
#define WINDOW_REGS    2
#define WINDOW_EEPROM  3
#define WINDOW_WATCH   4
#define WINDOW_PORTS   5
#define WINDOW_WHERE   6
#define WINDOW_DISASM  7
#define NUM_WINDOWS    8
#define WINDOW_DISPLAYED 0x01
#define WINDOW_SCROLLS   0x02
#define HORIZ_BAR        '\304'   /* the char '' */
#define VERT_BAR         '\263'   /* the char '' */
#define LEFT_JOIN        '\264'   /* the char '' */
#define RIGHT_JOIN       '\303'   /* the char '' */
#define BOTH_JOIN        '\305'   /* the char '' */
struct windata
   {
   WORD startrow;  /* starting row of displayed data, if != 0 header line is at startrow-1 */
   WORD startcol;  /* starting column of displayed data */
   WORD numrows;   /* number of rows in window */
   WORD numcols;   /* number of cols in window */
   WORD flags;     /* attributes of this window */
   WORD startnum;  /* internal use for window control code */
   WORD endnum;    /* internal use for window control code */
   BYTE startchar; /* how this header line starts */
   BYTE endchar;   /* how this header line ends */
   void (*scrollfn)();  /* scroll this function */
   };

#ifdef PICTYPE_16F877
   #define CORE_14BIT
   #define HAS_UART
   #ifdef HAS_UART
      #define TX_TRIS TRISC
      #define TX_PIN NOT_BIT6
   #endif  /* HAS_UART */
   #define HAS_ADC
   #ifdef HAS_ADC
      #define ADC_10BITS  /* else 8-bit ADC */
      #define HAS_ADCON1
   #endif  /* HAS_ADC */
   #define HAS_CCP
   #ifdef HAS_CCP
      #define CCP1_PORT PORTC
      #define CCP1_PIN BIT2
      #define NOT_CCP1_PIN NOT_BIT2
      #define HAS_CCP2
   #endif  /* HAS_CCP */
   #define HAS_TMR0
   #define HAS_TMR1
   #define HAS_TMR2
   #define HAS_IRP
   #define HAS_PORTB
   #define HAS_PORTC
   #define HAS_PORTD
   #define HAS_PORTE
   #define HAS_SSP
   #define HAS_PCON
   #define HAS_PIR1
   #define HAS_PIR2
   #define SELF_FLASH
   #define HAS_CPD
   #ifdef HAS_PCON
   /* #define HAS_OSCF */
   #endif  /* HAS_PCON */
   #define MEMORY_SIZE 8192
   #define MEMORY_MASK 8191   /* 0x1fff */
   #define STATUS_MEMORYPAGE_BITMASK 0x60  /* (STATUS_RP0|STATUS_RP1) */  /* 0110 0000 : bits 6 .. 5 */
   #define STATUS_MEMORYPAGE_BITSHIFT 2    /* to move page bits to high byte of word */
   #define NUM_REGS 512                /* registers addressable */
   #define HIDDEN_REGS 0               /* keep this a multiple of 4 */
   #define EEPROM_SIZE 256
   #define HAS_CP_BITS
   #ifdef HAS_CP_BITS
      #define CP00_MIN    0x0000
      #define CP00_MAX    0x1fff
      #define CP01_MIN    0x1000
      #define CP01_MAX    0x1fff
      #define CP10_MIN    0x1f00
      #define CP10_MAX    0x1fff
      #define CP11_MIN    0xffff  /* essentially OFF */
      #define CP11_MAX    0xffff  /* essentially OFF */
      #define CONFIG_CPLOW   0x0030       /* Code Protection bits (lowh copy) */
      #define CPLOW_BITS_SHIFT  4         /* turn CPLOW bits into an index */
      #define CONFIG_CPHIGH  0x3000       /* Code Protection bits (high copy) */
      #define CPHIGH_BITS_SHIFT 12        /* turn CPHIGH bits into an index */
   #endif  /* HAS_CP_BITS */
   #define ID_LOC      0x2000
   #define ID_END      0x2003
   #define CONFIG_LOC  0x2007
   #define CONFIG_MASK 0x3fff
   #define CONFIG_WDTE 0x0004          /* WDTE bit */
   #define EEPROM_LOC  0x2100  /* PICs put EEPROM info here */
/* #define PINS_8 */             /* limited I/O pins available */
   #define NUM_BYTE_PORTS 5      /* A, B, C, D, E */
   #define NUM_BYTE_PORTS_ALLOCATE 8  /* a multiple of 4 for data alignment */
   #define NUM_PORT_BITS (6+8+8+8+3)  /* A+B+C+D+E */
   #define ADC_CONFIG_MASK 0x0f
/* #define USE_ADDRLIST */
   #include <p877regd.h>
   #ifdef PSHELL   /* if it's time to declare stuff */
      struct windata window_data[NUM_WINDOWS] =
         { /* sr  sc  nr  nc  flags                           sn  en  startchar  endchar */
            { 25,  0, 25, 80, WINDOW_DISPLAYED,                0,  0, HORIZ_BAR, LEFT_JOIN, 0 },  /* command window */
            {  0,  0,  6, 80, WINDOW_DISPLAYED,                0,  0, HORIZ_BAR, HORIZ_BAR, 0 },  /* instr window */
            {  7,  0,  8, 80, WINDOW_DISPLAYED|WINDOW_SCROLLS, 0,127, HORIZ_BAR, LEFT_JOIN, scroll_regs },  /* regs window */
            { 16,  0,  4, 80, WINDOW_DISPLAYED|WINDOW_SCROLLS, 0, 63, HORIZ_BAR, LEFT_JOIN, scroll_eeprom },  /* eeprom window */
            { 21,  0,  3, 80, WINDOW_DISPLAYED,                0,  0, HORIZ_BAR, LEFT_JOIN, 0 },  /* watch window */
            {  1, 80, 11, 51, WINDOW_DISPLAYED,                0,  0, RIGHT_JOIN, HORIZ_BAR, 0 },  /* ports window */
            { 13, 80, 16, 51, WINDOW_DISPLAYED|WINDOW_SCROLLS,16,  0, RIGHT_JOIN, HORIZ_BAR, scroll_where },  /* where window */
            { 30, 80, 20, 51, WINDOW_DISPLAYED|WINDOW_SCROLLS, 0, 20, RIGHT_JOIN, HORIZ_BAR, scroll_disasm }  /* disassembly window */
         };
      struct windata vga_data[NUM_WINDOWS] =
         { /* sr  sc  nr  nc  flags                           sn  en  startchar  endchar */
            { 25,  0, 25, 79, WINDOW_DISPLAYED,                0,  0, HORIZ_BAR, HORIZ_BAR, 0 },  /* command window */
            {  0,  0,  6, 79, WINDOW_DISPLAYED,                0,  0, HORIZ_BAR, HORIZ_BAR, 0 },  /* instr window */
            {  7,  0,  8, 79, WINDOW_DISPLAYED|WINDOW_SCROLLS, 0,127, HORIZ_BAR, HORIZ_BAR, scroll_regs },  /* regs window */
            { 16,  0,  4, 79, WINDOW_DISPLAYED|WINDOW_SCROLLS, 0, 63, HORIZ_BAR, HORIZ_BAR, scroll_eeprom },  /* eeprom window */
            { 21,  0,  3, 79, WINDOW_DISPLAYED,                0,  0, HORIZ_BAR, HORIZ_BAR, 0 },  /* watch window */
            { 31,  0,  0,  0, 0,                               0,  0, RIGHT_JOIN, HORIZ_BAR, 0 },  /* ports window */
            {  0,  0,  0,  0, 0,                               0,  0, RIGHT_JOIN, HORIZ_BAR, 0 },  /* where window */
            {  0,  0,  0,  0, 0,                               0,  0, BOTH_JOIN, HORIZ_BAR, 0 }  /* disassembly window */
         };
   #endif  /* PSHELL */
#endif  /* PICTYPE_16F877 */

#ifdef PICTYPE_16F648
   #define CORE_14BIT
   #define HAS_UART
   #ifdef HAS_UART
      #define TX_TRIS TRISB
      #define TX_PIN NOT_BIT2
   #endif  /* HAS_UART */
/* #define HAS_ADC */
   #ifdef HAS_ADC
      #define ADC_10BITS  /* else 8-bit ADC */
/*    #define HAS_ADCON1 */
   #endif  /* HAS_ADC */
   #define HAS_CCP
   #ifdef HAS_CCP
      #define CCP1_PORT PORTB
      #define CCP1_PIN BIT3
      #define NOT_CCP1_PIN NOT_BIT3
/*    #define HAS_CCP2 */
   #endif  /* HAS_CCP */
   #define HAS_TMR0
   #define HAS_TMR1
   #define HAS_TMR2
   #define HAS_IRP
   #define HAS_PORTB
/* #define HAS_PORTC */
/* #define HAS_PORTD */
/* #define HAS_PORTE */
/* #define HAS_SSP */
   #define HAS_PCON
   #define HAS_PIR1
/* #define HAS_PIR2 */
/* #define SELF_FLASH */
   #define HAS_CPD
   #ifdef HAS_PCON
      #define HAS_OSCF
   #endif  /* HAS_PCON */
   #define HAS_CMCON
   #ifdef HAS_CMCON
/*    #define CMCON_AT_0X19 */
      #define CMCON_CONFIG_MASK 0x07
      #define HAS_CMP2
   #endif  /* HAS_CMCON */
   #define HAS_VRCON
   #ifdef HAS_VRCON
      #define VREF_PIN 2  /* A2 */
   #endif  /* HAS_VRCON */
   #define MEMORY_SIZE 4096
   #define MEMORY_MASK 4095   /* 0x0fff */
   #define STATUS_MEMORYPAGE_BITMASK 0x60  /* (STATUS_RP0|STATUS_RP1) */  /* 0110 0000 : bits 6 .. 5 */
   #define STATUS_MEMORYPAGE_BITSHIFT 2    /* to move page bits to high byte of word */
   #define NUM_REGS 512                /* registers addressable */
   #define HIDDEN_REGS 0               /* keep this a multiple of 4 */
   #define EEPROM_SIZE 256
   #define HAS_CP_BITS
   #ifdef HAS_CP_BITS
      #define CP00_MIN    0x0000
      #define CP00_MAX    0x07ff
      #define CP01_MIN    0x0200
      #define CP01_MAX    0x07ff
      #define CP10_MIN    0x0400
      #define CP10_MAX    0x07ff
      #define CP11_MIN    0xffff  /* essentially OFF */
      #define CP11_MAX    0xffff  /* essentially OFF */
      #define CONFIG_CPLOW   0x0e00       /* Code Protection bits (lowh copy) */
      #define CPLOW_BITS_SHIFT  10        /* turn CPLOW bits into an index */
      #define CONFIG_CPHIGH  0x3000       /* Code Protection bits (high copy) */
      #define CPHIGH_BITS_SHIFT 12        /* turn CPHIGH bits into an index */
   #endif  /* HAS_CP_BITS */
   #define ID_LOC      0x2000
   #define ID_END      0x2003
   #define CONFIG_LOC  0x2007
   #define CONFIG_MASK 0x3fff
   #define CONFIG_MCLRE 0x20           /* BIT5 */
   #define CONFIG_LVP   0x80           /* BIT7 */
   #define CONFIG_FOSC2 0x10           /* BIT4 */
   #define CONFIG_FOSC2_SHIFT 2        /* make FOSC2:FOSC1:FOSC0 continguous */
   #define CONFIG_FOSC_MASK 0x03       /* (BIT1|BIT0) */
   #define CONFIG_WDTE 0x0004          /* WDTE bit */
   #define EEPROM_LOC  0x2100  /* PICs put EEPROM info here */
/* #define PINS_8 */             /* limited I/O pins available */
   #define NUM_BYTE_PORTS 2      /* A, B */
   #define NUM_BYTE_PORTS_ALLOCATE 4  /* a multiple of 4 for data alignment */
   #define NUM_PORT_BITS (8+8)  /* A+B */
/* #define ADC_CONFIG_MASK 0x0f */
/* #define USE_ADDRLIST */
   #include <p648regd.h>
   #ifdef PSHELL   /* if it's time to declare stuff */
      struct windata window_data[NUM_WINDOWS] =
         { /* sr  sc  nr  nc  flags                           sn  en  startchar  endchar */
            { 25,  0, 25, 80, WINDOW_DISPLAYED,                0,  0, HORIZ_BAR, BOTH_JOIN, 0 },  /* command window */
            {  0,  0,  6, 80, WINDOW_DISPLAYED,                0,  0, HORIZ_BAR, HORIZ_BAR, 0 },  /* instr window */
            {  7,  0,  8, 80, WINDOW_DISPLAYED|WINDOW_SCROLLS, 0,127, HORIZ_BAR, LEFT_JOIN, scroll_regs },  /* regs window */
            { 16,  0,  4, 80, WINDOW_DISPLAYED|WINDOW_SCROLLS, 0, 63, HORIZ_BAR, LEFT_JOIN, scroll_eeprom },  /* eeprom window */
            { 21,  0,  3, 80, WINDOW_DISPLAYED,                0,  0, HORIZ_BAR, LEFT_JOIN, 0 },  /* watch window */
            {  1, 80,  4, 51, WINDOW_DISPLAYED,                0,  0, RIGHT_JOIN, HORIZ_BAR, 0 },  /* ports window */
            {  7, 80, 17, 51, WINDOW_DISPLAYED|WINDOW_SCROLLS,16,  0, BOTH_JOIN, HORIZ_BAR, scroll_where },  /* where window */
            { 25, 80, 25, 51, WINDOW_DISPLAYED|WINDOW_SCROLLS, 0, 25, BOTH_JOIN, HORIZ_BAR, scroll_disasm }  /* disassembly window */
         };
      struct windata vga_data[NUM_WINDOWS] =
         { /* sr  sc  nr  nc  flags                           sn  en  startchar  endchar */
            { 25,  0, 25, 79, WINDOW_DISPLAYED,                0,  0, HORIZ_BAR, HORIZ_BAR, 0 },  /* command window */
            {  0,  0,  6, 79, WINDOW_DISPLAYED,                0,  0, HORIZ_BAR, HORIZ_BAR, 0 },  /* instr window */
            {  7,  0,  8, 79, WINDOW_DISPLAYED|WINDOW_SCROLLS, 0,127, HORIZ_BAR, HORIZ_BAR, scroll_regs },  /* regs window */
            { 16,  0,  4, 79, WINDOW_DISPLAYED|WINDOW_SCROLLS, 0, 63, HORIZ_BAR, HORIZ_BAR, scroll_eeprom },  /* eeprom window */
            { 21,  0,  3, 79, WINDOW_DISPLAYED,                0,  0, HORIZ_BAR, HORIZ_BAR, 0 },  /* watch window */
            { 39,  0,  0,  0, 0,                               0,  0, RIGHT_JOIN, HORIZ_BAR, 0 },  /* ports window */
            {  0,  0,  0,  0, 0,                               0,  0, RIGHT_JOIN, HORIZ_BAR, 0 },  /* where window */
            {  0,  0,  0,  0, 0,                               0,  0, BOTH_JOIN, HORIZ_BAR, 0 }  /* disassembly window */
         };
   #endif  /* PSHELL */
#endif  /* PICTYPE_16F648 */

#ifdef PICTYPE_12F675
   #define CORE_14BIT
/* #define HAS_UART */
   #ifdef HAS_UART
      #define TX_TRIS TRISB
      #define TX_PIN NOT_BIT2
   #endif  /* HAS_UART */
   #define HAS_ADC
   #ifdef HAS_ADC
      #define ADC_10BITS  /* else 8-bit ADC */
/*    #define HAS_ADCON1 */
   #endif  /* HAS_ADC */
/* #define HAS_CCP */
   #ifdef HAS_CCP
      #define CCP1_PORT PORTB
      #define CCP1_PIN BIT3
      #define NOT_CCP1_PIN NOT_BIT3
/*    #define HAS_CCP2 */
   #endif  /* HAS_CCP */
   #define HAS_TMR0
   #define HAS_TMR1
/* #define HAS_TMR2 */
/* #define HAS_IRP */
/* #define HAS_PORTB */
/* #define HAS_PORTC */
/* #define HAS_PORTD */
/* #define HAS_PORTE */
/* #define HAS_SSP */
   #define HAS_PCON
   #define HAS_PIR1
/* #define HAS_PIR2 */
/* #define SELF_FLASH */
   #define HAS_CPD
   #ifdef HAS_PCON
   /* #define HAS_OSCF */
   #endif  /* HAS_PCON */
   #define HAS_CMCON
   #ifdef HAS_CMCON
      #define CMCON_AT_0X19
      #define CMCON_CONFIG_MASK 0x07
/*    #define HAS_CMP2 */
   #endif  /* HAS_CMCON */
   #define HAS_VRCON
   #define HAS_OSCCAL
   #define HAS_WPU
   #define HAS_IOC
   #define HAS_ANSEL
   #define HAS_TMR1GE
   #define MEMORY_SIZE 1024
   #define MEMORY_MASK 1023   /* 0x03ff */
   #define STATUS_MEMORYPAGE_BITMASK STATUS_RP0  /* 0010 0000 : bit 5 */
   #define STATUS_MEMORYPAGE_BITSHIFT 2    /* to move page bits to high byte of word */
   #define NUM_REGS 256                /* registers addressable */
   #define HIDDEN_REGS 0               /* keep this a multiple of 4 */
   #define EEPROM_SIZE 128
/* #define HAS_CP_BITS */
   #ifdef HAS_CP_BITS
      #define CP00_MIN    0x0000
      #define CP00_MAX    0x07ff
      #define CP01_MIN    0x0200
      #define CP01_MAX    0x07ff
      #define CP10_MIN    0x0400
      #define CP10_MAX    0x07ff
      #define CP11_MIN    0xffff  /* essentially OFF */
      #define CP11_MAX    0xffff  /* essentially OFF */
      #define CONFIG_CPLOW   0x0e00       /* Code Protection bits (lowh copy) */
      #define CPLOW_BITS_SHIFT  10        /* turn CPLOW bits into an index */
      #define CONFIG_CPHIGH  0x3000       /* Code Protection bits (high copy) */
      #define CPHIGH_BITS_SHIFT 12        /* turn CPHIGH bits into an index */
   #endif  /* HAS_CP_BITS */
   #define ID_LOC      0x2000
   #define ID_END      0x2003
   #define CONFIG_LOC  0x2007
   #define CONFIG_MASK 0x3fff
   #define CONFIG_MCLRE 0x20           /* BIT5 */
   #define CONFIG_FOSC_MASK 0x07       /* (BIT2|BIT1|BIT0) */
   #define CONFIG_WDTE 0x0008          /* WDTE bit */
   #define EEPROM_LOC  0x2100  /* PICs put EEPROM info here */
   #define PINS_8                /* limited I/O pins available */
   #define NUM_BYTE_PORTS 1      /* A */
   #define NUM_BYTE_PORTS_ALLOCATE 4  /* a multiple of 4 for data alignment */
   #define NUM_PORT_BITS 6
   #define ADC_CONFIG_MASK 0x0f
   #define OSCCAL_INSTR 0x34f0         /* RETLW 0f0h */
   #define OSCCAL_INSTR_LOC (MEMORY_SIZE-1)
   #define RESET_OSCCAL 0x80           /* 1000 00--b */
/* #define USE_ADDRLIST */
   #include <p675regd.h>
   #ifdef PSHELL   /* if it's time to declare stuff */
      struct windata window_data[NUM_WINDOWS] =
         { /* sr  sc  nr  nc  flags                           sn  en  startchar  endchar */
            { 25,  0, 25, 80, WINDOW_DISPLAYED,                0,  0, HORIZ_BAR, BOTH_JOIN, 0 },  /* command window */
            {  0,  0,  6, 80, WINDOW_DISPLAYED,                0,  0, HORIZ_BAR, HORIZ_BAR, 0 },  /* instr window */
            {  7,  0,  8, 80, WINDOW_DISPLAYED|WINDOW_SCROLLS, 0,127, HORIZ_BAR, LEFT_JOIN, scroll_regs },  /* regs window */
            { 16,  0,  4, 80, WINDOW_DISPLAYED|WINDOW_SCROLLS, 0, 63, HORIZ_BAR, LEFT_JOIN, scroll_eeprom },  /* eeprom window */
            { 21,  0,  3, 80, WINDOW_DISPLAYED,                0,  0, HORIZ_BAR, LEFT_JOIN, 0 },  /* watch window */
            {  1, 80,  4, 51, WINDOW_DISPLAYED,                0,  0, RIGHT_JOIN, HORIZ_BAR, 0 },  /* ports window */
            {  5, 80, 19, 51, WINDOW_DISPLAYED|WINDOW_SCROLLS,16,  0, RIGHT_JOIN, HORIZ_BAR, scroll_where },  /* where window */
            { 25, 80, 25, 51, WINDOW_DISPLAYED|WINDOW_SCROLLS, 0, 25, BOTH_JOIN, HORIZ_BAR, scroll_disasm }  /* disassembly window */
         };
      struct windata vga_data[NUM_WINDOWS] =
         { /* sr  sc  nr  nc  flags                           sn  en  startchar  endchar */
            { 25,  0, 25, 79, WINDOW_DISPLAYED,                0,  0, HORIZ_BAR, HORIZ_BAR, 0 },  /* command window */
            {  0,  0,  6, 79, WINDOW_DISPLAYED,                0,  0, HORIZ_BAR, HORIZ_BAR, 0 },  /* instr window */
            {  7,  0,  8, 79, WINDOW_DISPLAYED|WINDOW_SCROLLS, 0,127, HORIZ_BAR, HORIZ_BAR, scroll_regs },  /* regs window */
            { 16,  0,  4, 79, WINDOW_DISPLAYED|WINDOW_SCROLLS, 0, 63, HORIZ_BAR, HORIZ_BAR, scroll_eeprom },  /* eeprom window */
            { 21,  0,  3, 79, WINDOW_DISPLAYED,                0,  0, HORIZ_BAR, HORIZ_BAR, 0 },  /* watch window */
            { 39,  0,  0,  0, 0,                               0,  0, RIGHT_JOIN, HORIZ_BAR, 0 },  /* ports window */
            {  0,  0,  0,  0, 0,                               0,  0, RIGHT_JOIN, HORIZ_BAR, 0 },  /* where window */
            {  0,  0,  0,  0, 0,                               0,  0, BOTH_JOIN, HORIZ_BAR, 0 }  /* disassembly window */
         };
   #endif  /* PSHELL */
#endif  /* PICTYPE_12F675 */

#ifdef PICTYPE_16F84A
   #define CORE_14BIT
/* #define HAS_UART */
   #ifdef HAS_UART
      #define TX_TRIS TRISB
      #define TX_PIN NOT_BIT2
   #endif  /* HAS_UART */
/* #define HAS_ADC */
   #ifdef HAS_ADC
      #define ADC_10BITS  /* else 8-bit ADC */
/*    #define HAS_ADCON1 */
   #endif  /* HAS_ADC */
/* #define HAS_CCP */
   #ifdef HAS_CCP
      #define CCP1_PORT PORTB
      #define CCP1_PIN BIT3
      #define NOT_CCP1_PIN NOT_BIT3
/*    #define HAS_CCP2 */
   #endif  /* HAS_CCP */
   #define HAS_TMR0
/* #define HAS_TMR1 */
/* #define HAS_TMR2 */
/* #define HAS_IRP */
   #define HAS_PORTB
/* #define HAS_PORTC */
/* #define HAS_PORTD */
/* #define HAS_PORTE */
/* #define HAS_SSP */
/* #define HAS_PCON */
/* #define HAS_PIR1 */
   #ifdef HAS_PIR1
/*    #define HAS_PIR2 */
   #endif  /* HAS_PIR1 */
/* #define SELF_FLASH */
   #define HAS_CPD
   #ifdef HAS_PCON
      #define HAS_OSCF
   #endif  /* HAS_PCON */
/* #define HAS_CMCON */
   #ifdef HAS_CMCON
/*    #define CMCON_AT_0X19 */
      #define CMCON_CONFIG_MASK 0x07
      #define HAS_CMP2
   #endif  /* HAS_CMCON */
/* #define HAS_VRCON */
   #ifdef HAS_VRCON
      #define VREF_PIN 2  /* A2 */
   #endif  /* HAS_VRCON */
   #define MEMORY_SIZE 1024
   #define MEMORY_MASK 1023   /* 0x1fff */
   #define STATUS_MEMORYPAGE_BITMASK STATUS_RP0  /* 0010 0000 : bit 5 */
   #define STATUS_MEMORYPAGE_BITSHIFT 2    /* to move page bits to high byte of word */
   #define NUM_REGS 256                /* registers addressable */
   #define HIDDEN_REGS 0               /* keep this a multiple of 4 */
   #define EEPROM_SIZE 64
   #ifdef EEPROM_SIZE
      #define EEIF_IN_EECON1
   #endif  /* EEPROM_SIZE */
   #define CP00_MIN    0xffff  /* essentially OFF */
   #define CP00_MAX    0xffff  /* essentially OFF */
   #define CP01_MIN    0xffff  /* essentially OFF */
   #define CP01_MAX    0xffff  /* essentially OFF */
   #define CP10_MIN    0xffff  /* essentially OFF */
   #define CP10_MAX    0xffff  /* essentially OFF */
   #define CP11_MIN    0xffff  /* essentially OFF */
   #define CP11_MAX    0xffff  /* essentially OFF */
   #define ID_LOC      0x2000
   #define ID_END      0x2003
   #define CONFIG_LOC  0x2007
   #define CONFIG_MASK 0x3fff
   #define CONFIG_MCLRE 0x020          /* BIT5 */
   #define CONFIG_LVP   0x080          /* BIT7 */
   #define CONFIG_WDTE 0x0004          /* WDTE bit */
   #define EEPROM_LOC  0x2100  /* PICs put EEPROM info here */
/* #define PINS_8 */             /* limited I/O pins available */
   #define NUM_BYTE_PORTS 2      /* A, B */
   #define NUM_BYTE_PORTS_ALLOCATE 4  /* a multiple of 4 for data alignment */
   #define NUM_PORT_BITS (6+8)  /* A+B */
/* #define USE_ADDRLIST */
   #include <p84regd.h>
   #ifdef PSHELL   /* if it's time to declare stuff */
      struct windata window_data[NUM_WINDOWS] =
         { /* sr  sc  nr  nc  flags                           sn  en  startchar  endchar */
            { 25,  0, 25, 80, WINDOW_DISPLAYED,                0,  0, HORIZ_BAR, BOTH_JOIN, 0 },  /* command window */
            {  0,  0,  6, 80, WINDOW_DISPLAYED,                0,  0, HORIZ_BAR, HORIZ_BAR, 0 },  /* instr window */
            {  7,  0,  8, 80, WINDOW_DISPLAYED|WINDOW_SCROLLS, 0,127, HORIZ_BAR, LEFT_JOIN, scroll_regs },  /* regs window */
            { 16,  0,  4, 80, WINDOW_DISPLAYED,                0, 63, HORIZ_BAR, LEFT_JOIN, 0 },  /* eeprom window */
            { 21,  0,  3, 80, WINDOW_DISPLAYED,                0,  0, HORIZ_BAR, LEFT_JOIN, 0 },  /* watch window */
            {  1, 80,  4, 51, WINDOW_DISPLAYED,                0,  0, RIGHT_JOIN, HORIZ_BAR, 0 },  /* ports window */
            {  7, 80, 17, 51, WINDOW_DISPLAYED|WINDOW_SCROLLS,16,  0, BOTH_JOIN, HORIZ_BAR, scroll_where },  /* where window */
            { 25, 80, 25, 51, WINDOW_DISPLAYED|WINDOW_SCROLLS, 0, 25, BOTH_JOIN, HORIZ_BAR, scroll_disasm }  /* disassembly window */
         };
      struct windata vga_data[NUM_WINDOWS] =
         { /* sr  sc  nr  nc  flags                           sn  en  startchar  endchar */
            { 25,  0, 25, 79, WINDOW_DISPLAYED,                0,  0, HORIZ_BAR, HORIZ_BAR, 0 },  /* command window */
            {  0,  0,  6, 79, WINDOW_DISPLAYED,                0,  0, HORIZ_BAR, HORIZ_BAR, 0 },  /* instr window */
            {  7,  0,  8, 79, WINDOW_DISPLAYED|WINDOW_SCROLLS, 0,127, HORIZ_BAR, HORIZ_BAR, scroll_regs },  /* regs window */
            { 16,  0,  4, 79, WINDOW_DISPLAYED,                0, 63, HORIZ_BAR, HORIZ_BAR, 0 },  /* eeprom window */
            { 21,  0,  3, 79, WINDOW_DISPLAYED,                0,  0, HORIZ_BAR, HORIZ_BAR, 0 },  /* watch window */
            { 39,  0,  0,  0, 0,                               0,  0, RIGHT_JOIN, HORIZ_BAR, 0 },  /* ports window */
            {  0,  0,  0,  0, 0,                               0,  0, RIGHT_JOIN, HORIZ_BAR, 0 },  /* where window */
            {  0,  0,  0,  0, 0,                               0,  0, BOTH_JOIN, HORIZ_BAR, 0 }  /* disassembly window */
         };
   #endif  /* PSHELL */
#endif  /* PICTYPE_16F84A */

#ifdef PICTYPE_12C509A
   #define CORE_12BIT
   #define HAS_TMR0
   #define HAS_OSCCAL
   #define MEMORY_SIZE 1024
   #define MEMORY_MASK 1023  /* 0x03ff */
   #define HIDDEN_REGS 4               /* keep this a multiple of 4 */
   #define ID_LOC      0x0400
   #define ID_END      0x0403
   #define CONFIG_LOC  0x0fff
   #define CONFIG_MASK 0x001f
   #define WAKEPINS 0x0b               /* 1011 -- GP3, GP1, GP0 */
   #define WAKEPINS_NO_GP3 0x03        /* 0011 -- GP1, GP0 */
   #define PINS_8                /* limited I/O pins available */
/* #define USE_ADDRLIST */
   #include <p509regd.h>
   #ifdef PSHELL   /* if it's time to declare stuff */
      struct windata window_data[NUM_WINDOWS] =
         { /* sr  sc  nr  nc  flags                           sn  en  startchar  endchar */
            { 25,  0, 25, 80, WINDOW_DISPLAYED,                0,  0, HORIZ_BAR, LEFT_JOIN, 0 },  /* command window */
            {  0,  0,  6, 80, WINDOW_DISPLAYED,                0,  0, HORIZ_BAR, HORIZ_BAR, 0 },  /* instr window */
            {  7,  0,  4, 80, WINDOW_DISPLAYED,                0, 63, HORIZ_BAR, LEFT_JOIN, 0 },  /* regs window */
            {  0,  0,  0,  0, 0,                               0,  0, HORIZ_BAR, LEFT_JOIN, 0 },  /* eeprom window */
            { 21,  0,  3, 80, WINDOW_DISPLAYED,                0,  0, HORIZ_BAR, LEFT_JOIN, 0 },  /* watch window */
            { 12,  0,  2, 80, WINDOW_DISPLAYED,                0,  0, HORIZ_BAR, LEFT_JOIN, 0 },  /* ports window */
            {  1, 80, 16, 51, WINDOW_DISPLAYED|WINDOW_SCROLLS, 0,  0, RIGHT_JOIN, HORIZ_BAR, scroll_where },  /* where window */
            { 18, 80, 32, 51, WINDOW_DISPLAYED|WINDOW_SCROLLS,0x3ff,30,RIGHT_JOIN, HORIZ_BAR, scroll_disasm }  /* disassembly window */
         };
      struct windata vga_data[NUM_WINDOWS] =
         { /* sr  sc  nr  nc  flags                           sn  en  startchar  endchar */
            { 25,  0, 25, 79, WINDOW_DISPLAYED,                0,  0, HORIZ_BAR, HORIZ_BAR, 0 },  /* command window */
            {  0,  0,  6, 79, WINDOW_DISPLAYED,                0,  0, HORIZ_BAR, HORIZ_BAR, 0 },  /* instr window */
            {  7,  0,  4, 79, WINDOW_DISPLAYED,                0, 63, HORIZ_BAR, HORIZ_BAR, 0 },  /* regs window */
            {  0,  0,  0,  0, 0,                               0,  0, HORIZ_BAR, HORIZ_BAR, 0 },  /* eeprom window */
            { 21,  0,  3, 79, WINDOW_DISPLAYED,                0,  0, HORIZ_BAR, HORIZ_BAR, 0 },  /* watch window */
            { 12,  0,  2, 79, WINDOW_DISPLAYED,                0,  0, HORIZ_BAR, HORIZ_BAR, 0 },  /* ports window */
            {  0,  0,  0,  0, 0,                               0,  0, HORIZ_BAR, HORIZ_BAR, 0 },  /* where window */
            { 16,  0,  4, 79, WINDOW_DISPLAYED|WINDOW_SCROLLS,0x3ff,3,HORIZ_BAR,  HORIZ_BAR, scroll_disasm }  /* disassembly window */
         };
   #endif  /* PSHELL */
#endif  /* PICTYPE_12C509A */

#define MAX_MEMORY_SIZE 8192   /* max for PIC chip */
#define MAX_MEMORY_MASK 8191
#define MAX_STACK_SIZE 16
#define PIC_DECODE_BITS 6      /* use upper 6 bits for instruction decode */

#define STATUS_FLAGBITS 0x07
#define STATUS_OTHERBITS 0xf8


#ifdef CORE_14BIT
   #define MAX_NUM_REGS 512
   #define STACK_SIZE 8                /* stack size */
   #define STACK_MASK 7
   #define PCLATH_GOTOCALL_MASK 0x18  /* bits 3..4 are bits 11..12 of address for GOTO/CALL */
   #define PCLATH_BITSHIFT 8    /* must be moved 8 bits */
   #define OPTYPES_SHIFT 8             /* shift to move OPTYPES index to bottom of    #define */
   #define INSTRUCTION_BITMASK 0x3fff  /* 14-bit instructions */
   #define REGINDEX_MASK 0x007f        /* 7-bit register file mask */
   #define DEST_MASK 0x0080            /* isolate F/!W bit */
   #define BITNUM_MASK 0x0380          /* isolate bitnumber from instr */
   #define BITNUM_POSITION 7           /* shift to move bitnum to bottom of word */
   #define RESET_IP 0x0000             /* IP value at reset */
   #define RESET_FSR 0x00              /* 0000 0000b */
   #define RESET_STATUS 0x18           /* 0001 1000b */
   #define OPTION_RESET 0xff           /* option reg at reset */
   #define OPTION_PRESCALE_BITS 0x07   /* bits 0 .. 2 */
   #define CONFIG_WRT  0x0200          /* WRT bit (program memory write enable) */
   #define CONFIG_CPD 0x0100           /* CPD: Data EE memory code protection */
   #define FSR_MASK 0xff               /* maximum index into memory */
   #define GOTO_MASK 0x07ff  /* mask 11 bits of GOTO destination from opcode */
   #define CALL_MASK 0x07ff  /* mask 11 bits of CALL destination from opcode */
   #define NUM_INSTRS (37+4)           /* for the assembler, +4 for pseudo-ops */
      /* NOTE: The documentation says 35 instructions because they want to
       *       discourage the use of the OPTION and TRIS instructions.
       */
#endif  /* CORE_14BIT */

#ifdef CORE_12BIT
   #define MAX_NUM_REGS 64
   #define NUM_REGS 64                 /* registers available */
   #define NUM_BYTE_PORTS 1
   #define NUM_BYTE_PORTS_ALLOCATE 4  /* a multiple of 4 for data alignment */
   #define NUM_PORT_BITS 6
   #define STACK_SIZE 2
   #define STACK_MASK 1
   #define OPTION_PRESCALE_BITS 0x07  /* bits 0 .. 2 */
   #define STATUS_PA0_SHIFT 4  /* to put this at bit 9 */
   #define FSR_MEMORYPAGE_BITMASK 0x20  /* 0010 0000 : bit 5 */
   #define OPTYPES_SHIFT 6  /* shift to move OPTYPES index to bottom of word */
   #define INSTRUCTION_BITMASK 0x0fff  /* 12-bit instructions */
   #define REGINDEX_MASK 0x001f        /* 5-bit register file mask */
   #define DEST_MASK 0x0020            /* isolate F/!W bit */
   #define BITNUM_MASK 0x00e0          /* isolate bitnumber from instr */
   #define BITNUM_POSITION 5           /* shift to move bitnum to bottom of word */
   #define RESET_IP 0x03ff             /* IP on RESET */
   #define OSCCAL_INSTR 0x0cf0         /* MOVLW 0f0h */
   #define OSCCAL_INSTR_LOC (MEMORY_SIZE-1)
   #define RESET_FSR 0xc0              /* 1100 0000b */
   #define RESET_OSCCAL 0x80           /* 1000 00--b */
   #define RESET_TRIS 0x3f             /* 0011 1111b */
   #define RESET_STATUS 0x18           /* 0001 1000b */
   #define OPTION_RESET 0xff           /* option reg at reset */
   #define CONFIG_WDTE 0x0004          /* WDTE bit */
   #define FSR_MASK 0x3f               /* maximum index into memory */
   #define GOTO_MASK_12BIT 0x01ff  /* mask 9 bits of GOTO destination from opcode */
   #define CALL_MASK_12BIT 0x00ff  /* mask 8 bits of CALL destination from opcode */
   #define CONFIG_MCLRE 0x10           /* BIT4 */
   #define NUM_INSTRS (33+4)           /* for the assembler, +4 for pseudo-ops */
#endif  /* CORE_12BIT */

#define READ_USES_TRIS
#define SLOW_REGBREAK_CHECKS
   /*
    *  Note:  When checking register breakpoints, there are actually
    *         several register numbers that can be checked:
    *         the register number (filenum_temp), the translated
    *         register number (i.e. in the 16F877, registers 0xf0 - 0xff
    *         translate back to 0x70 - 0x7f) (xlate_regs[filenum_temp]),
    *         and indirect register addressing.
    *
    *         So, with a 16F877 if you want to put a register breakpoint on
    *         0x72, you would get a "break" on access to 0x72 AND 0xf2 AND
    *         0x172 AND 0x1f2.  But what if you want to check for a breakpoint
    *         only on 0x1f2?
    *
    *         When SLOW_REGBREAK_CHECKS is set, breakpoints are checked
    *         for on BOTH the register number and the translated register
    *         number (i.e. you can set a breakpoint on 0x1f2 and NOT 0x072).
    *         If it is not set, then breakpoints are only checked for on the
    *         translated register number (i.e. a breakpoint on 0x1f2 will
    *         never be "TRUE", but a breakpoint on 0x072 will break on 0x072,
    *         0x0f2, 0x0172, and 0x01f2).
    *
    *         In terms of indirect addressing, breakpoints are check for on
    *         both the INDF (register 0) and it's destination (via STATUS
    *         and FSR) regardless of the setting of SLOW_REGBREAK_CHECKS.
    *
    *         A breakpoint on FSR will only be checked for direct or
    *         indirect addressing of FSR itself, not the use of FSR in
    *         indirect addressing.
    */

#define CODFILE_FIRST_MAPSECTOR 0x01bb
#define CODFILE_LAST_MAPSECTOR 0x01bd
#define CODFILE_SYMTAB_START 0x01aa
#define CODFILE_SYMTAB_END 0x01ac
#define CODFILE_SECTOR_SIZE 0x0200
#define MAP_SIZE 512   /* bytes in map table */
#define SECTMAP_SIZE 512  /* bytes in sector map table */
#define MAX_SYMBOLS 4095
#define SYMTAB_ENTRY_SIZE 16
#define SYMTAB_FLAGBYTE_OFS 13
#define SYMTAB_VALUE_OFS 14
#define SYMBOL_SIZE 14  /* includeing '\0' */
#define SYMTAB_FLAGBYTE_FIRSTLABEL 0x10
#define SYMTAB_FLAGBYTE_LASTLABEL  0x2e

#define SOFTUART_DEFAULT 1  /* NONE  (0=screen/keyboard) */

#define STRING_SIZE 128

#define PUTS(s) myputs(s)

#define WHERE_SIZE 0x7ff   /* size of "where was I?" queue */

struct membreaks
   {
   WORD low;
   WORD high;
   };
struct regbreaks
   {
   WORD low;
   WORD high;
   BYTE type;  /* bitflags: 0x01 = read, 0x02 = write */
   BYTE have_value;  /* if TRUE, use value for breakpoint */
   BYTE value;
   BYTE mask;
   };
#define MAX_REG_BREAKS 20
#define MAX_MEM_BREAKS 20
#define MAX_STICKY_BREAKS 10
#define REGBREAK_READ  0x01
#define REGBREAK_WRITE 0x02

#define ESC '\033'
#define CR '\015'
#define LF '\012'
#define BS '\010'
#define TAB '\011'
#define F1  0x13B
#define F2  0x13C
#define F3  0x13D
#define F4  0x13E
#define F5  0x13F
#define F6  0x140
#define F7  0x141
#define F8  0x142
#define F9  0x143
#define F10 0x144
#define F11 0x185
#define F12 0x186
#define CTRL_TAB 0x194
#define UPARROW 0x148
#define DNARROW 0x150
#define LTARROW 0x14B
#define RTARROW 0x14D
#define PGUP 0x149
#define PGDN 0x151
#define HOME 0x147
#define END 0x14f
#define INS 0x152
#define DEL 0x153

#define FIRST_MACRO_KEY F1
#define LAST_MACRO_KEY F10
#define NUM_MACRO_KEYS (LAST_MACRO_KEY - FIRST_MACRO_KEY + 1)


#define KEYMODE_INVERT 0x0000
      /* when a key is tied to an input pin, an unpressed key is 1, and a
       * pressed key is 0
       */
#define KEYMODE_FOLLOW 0x0001
      /* when a key is tied to an input pin, an unpressed key is 0, and a
       * pressed key is 1
       */
#define KEYMODE_TOGGLE 0x0002
   /* when a key is tied to an input pin, each keypress will toggle the
    * pin's current analog value (i.e. value ^= 0xffff)
    */
#define KEYMODE_INCREASE 0x0003
   /* when a key is tied to an input pin, each keypress will increase
    * the corresponding analog value by 0x0100 to a maximum of 0xffff
    */
#define KEYMODE_DECREASE 0x0004
   /* when a key is tied to an input pin, each keypress will decrease
    * the corresponding analog value by 0x0100 to a minimum of 0x0000
    */
#define KEYMODE_UNUSED 0xffff
   /* this key_pins[] entry is not used */
#define MAX_KEYPINS 10
/*
 * Note:  Since we are allowing analog values, we might want to have more
 *        than one functionkey associated with each input (for example,
 *        increment and decrement the associated analog value).  Thus, it
 *        no longer makes sense to limit the number of keypins to the number
 *        of I/O pins.
 *
 *#ifdef PINS_8
 *   #define NUM_KEYPINS 6  {* use the function keys F1 .. F6 *}
 *#else
 */
   #define NUM_KEYPINS MAX_KEYPINS
     /* use only the normal, un-<shift/ctrl/alt> function keys F1 .. F10 */
/*
#endif
*/

    /* NOTE:
     *       The function keys have scancodes of
     *              0x3b .. 0x44 for F1  .. F10
     *              0x57 .. 0x58 for F11 .. F12
     *       when pressed, and
     *              0xbb .. 0xc4 for F1  .. F10
     *              0xd7 .. 0xd8 for F11 .. F12
     *       when released.
     *
     *       The keycodes in the keyboard buffer for the various
     *       F-key combinations are:
     *          Normal:
     *             0x3b00 .. 0x4400 for F1  .. F10
     *             0x8500 .. 0x8600 for F11 .. F12
     *          Shift:
     *             0x5400 .. 0x5d00 for SF1  .. SF10
     *             0x8700 .. 0x8800 for SF11 .. SF12
     *          Ctrl:
     *             0x5e00 .. 0x6700 for CF1  .. CF10
     *             0x8900 .. 0x8a00 for CF11 .. CF12
     *          Alt:
     *             0x6800 .. 0x7100 for AF1  .. AF10
     *             0x8b00 .. 0x8c00 for AF11 .. AF12
     *
     *       If you want to implement additional keys mapped to input pins,
     *       I suggest using the keyboard "shift" flags at (BYTE)0x40:0x17
     *       to modify the scancodes noted above.
     *
     *       Beware, however, of changing shift flags between key press and
     *       release.  i.e. in a sequence like
     *                 <SHIFT><F1><Release SHIFT><Release F1>, the
     *       release of F1 should modify the pin associated with SF1, not
     *       F1.
     */
struct keypins
   {
   WORD keytype;
   void (*keyaction)();
   BYTE portnum;
   BYTE portbit;
   };

/*  Note:  Even though these are only used for hardware UART emulation,
 *         define them anyway so that only 1 version of take_int9() is
 *         needed.
 */
#define UART_QUEUE_SIZE 32   /* chars in IN/OUT buf, keep this a power of 2 */
#define UART_QUEUE_MASK 31
#define SCREEN_ROWS 25
#define SCREEN_COLS 80
#define MAX_SCREEN_WIDTH 132  /* 132 cols in VESA mode */

#ifdef HAS_UART
   #define WINNT_INT_DELAY 0xffff
   #define UART_FREQUENCY 115200L   /* frequency IN to UART */

   /* 8250 registers */

   #define UART_REC           0  /* UART receive reg */

   #define UART_XMIT          0  /* UART transmit reg */

   #define UART_INT_EN        1  /* UART int. enable reg */

   #define UART_INT_ID        2  /* UART int. ident. reg */

   #define UART_LINE_CONTROL  3  /* UART line control reg */

   #define UART_MODEM_CONTROL 4  /* UART modem control reg */

   #define UART_LINE_STATUS   5  /* UART line status reg */

   #define UART_MODEM_STATUS  6  /* UART modem status reg */

   #define UART_BAUD_LSB      0  /* UART baud divisor reg */

   #define UART_BAUD_MSB      1  /* UART baud divisor reg */
   #define UART_FCR_REG       2  /* FCR reg for 16550+ */


   #define UART_NONE          0  /* Handshake param none */

   #define UART_HDW           1  /* Handshake param hardware */

   #define UART_XON           2  /* Handshake param software */



   /* Interrupt enable register  */

   #define UART_RX_INT    0x01  /* Receive interrupt mask */

   #define UART_TBE_INT   0x02  /* Transmit buffer empty mask */

   #define UART_ERR_INT   0x04  /* Error interrupt mask */

   #define UART_RS_INT    0x08  /* Line interrupt mask */



   /* Interrupt id register */

   #define UART_OUT2      0x08  /* Out 2 line */

   #define UART_DTR       0x01  /* DTR high */

   #define UART_RTS       0x02  /* RTS high */

   #define UART_CTS       0x10

   #define UART_DSR       0x20

   #define UART_XMTRDY    0x20

   #define UART_TXR          0  /*  Transmit register (WRITE) */

#endif  /* HAS_UART */

#ifdef HAS_ADC
   #ifdef PICTYPE_16F877
      #define ADON_GODONE 0x05  /* (ADON | GODONE) */
      #define ADC_INTERNAL_RC 0xc0  /* (ADCS0|ADCS1) */
   #endif  /* PICTYPE_16F877 */
   #ifdef PICTYPE_12F675
      #define ADON_GODONE 0x03  /* (ADON | GODONE) */
      #define ADC_INTERNAL_RC 0x30  /* (ADCS0|ADCS1) both set */
   #endif  /* PICTYPE_12F675 */
   #define ADC_TABLE_MAXSIZE 2048  /* keep this a multiple of 2 */
#endif  /* HAS_ADC */

#ifdef HAS_CCP
   #define CCP_MODEMASK         0x0f
   #define CCP_MODE_SHIFT          2
   #define CCP_PWM_LOWBITS_MASK 0x30
   #define CCPMODE_DISABLE      0x00   /* NOTE: CCPMODE_xxxx is (CCPxCON >> CCP_MODE_SHIFT) & 3 */
   #define CCPMODE_CAPTURE      0x01
   #define CCPMODE_COMPARE      0x02
   #define CCPMODE_PWM          0x03
   #define CCPFN_DISABLE        0x00  /* NOTE: CCPFN_xxxx is CCPxCON & CCP_MODEMASK */
   #define CCPFN_FALLING        0x04
   #define CCPFN_RISING         0x05
   #define CCPFN_RISING4        0x06
   #define CCPFN_RISING16       0x07
   #define CCPFN_SETMATCH       0x08
   #define CCPFN_CLEARMATCH     0x09
   #define CCPFN_CMPAREINT      0x0a
   #define CCPFN_SPECIALEVENT   0x0b
   #define CCPFN_PWMMODE        0x0c
#endif  /* HAS_CCP */

#ifdef HAS_TMR1
   #define T1CON_ON_CHECK  0x03 /* (TMR1CS|TMR1ON) */ /* check TMR1CS, TMR1ON */
   #define T1_ON_VALUE 0x01  /* want TMR1CS==0 (FOSC/4), TMR1ON==1 */
   #define T1CON_PRESCALE_BITS  0x30  /* (T1CKPS0|T1CKPS1) */  /* bits 4..5 */
   #define T1CON_PRESCALE_SHIFT 4  /* move prescale bits to bottom of byte */
   #define T1OSCEN_TMR1CS 0x0a  /* (T1OSCEN|TMR1CS) */ /* check 3 bits */
   #define T1OSCEN_OFF_TMR1CS_ON 0x02  /* T1OSCEN clear, TMR1CS set for T1CKI update of Timer1 */
   #define T1OSCEN_ON_TMR1CS_ON  0x0a  /* T1OSCEN set, TMR1CS set for T1OSI update of Timer1 */
#endif  /* HAS_TMR1 */

#ifdef HAS_TMR2
   #define T2CON_PS_BITS 0x03  /* (T2CKPS0|T2CKPS1) */ /* low 2 bits */
   #define T2CON_POSTSCALE 0x78  /* (TOUTPS0|TOUTPS1|TOUTPS2|TOUTPS3) */  /* 4 bits of postscalar */
   #define T2CON_POSTSCALE_SHIFT 3
#endif  /* HAS_TMR2 */

#define BIT0     0x01
#define BIT1     0x02
#define BIT2     0x04
#define BIT3     0x08
#define BIT4     0x10
#define BIT5     0x20
#define BIT6     0x40
#define BIT7     0x80
#define NOT_BIT0 0xfe
#define NOT_BIT1 0xfd
#define NOT_BIT2 0xfb
#define NOT_BIT3 0xf7
#define NOT_BIT4 0xef
#define NOT_BIT5 0xdf
#define NOT_BIT6 0xbf
#define NOT_BIT7 0x7f

/*
 *  Define's for the assembler
 */
#define ASM_DATA    0xfffd   /* *(WORD *)browse_screen = number, *(WORD *)&browse_screen[2]=data, ... */
#define ASM_ORG     0xfffe   /* new org is in *(WORD *)browse_screen */
#define ASM_INVALID 0xffff   /* invalid opcode for error return */

#define ASM_NONE    0x0000
#define ASM_FILENUM 0x0001
#define ASM_W       0x0002
#define ASM_BITNUM  0x0004
#define ASM_8BITK   0x0008
#define ASM_9BITK   0x0010  /* 12-bit only */
#define ASM_11BITK  0x0020  /* 14-bit only */
#define ASM_TRIS14  0x0040  /* 14-bit only */
#define ASM_ORIGIN  0x0080  /* pseudo-op */
#define ASM_STORAGE 0x0100  /* generic storage */
#define ASM_DW      0x0200  /* pseudo-op */
#define ASM_DT      0x0400  /* pseudo-op */

#define BRK_NONE         0   /* no breakpoint */
#define BRK_KEYPRESS     1   /* Keypress break (Ctrl-Esc) */
#define BRK_MEMORY       2   /* memory location */
#define BRK_REGISTER     3   /* regsiter location */
#define BRK_COUNT        4   /* instruction count */
#define BRK_TIME         5   /* time expired */
#define BRK_STACKOVF     6   /* stack overflow */
#define BRK_STACKUNDF    7   /* stack underflow */
#define BRK_WATCHDOG     8   /* watchdog expired */
#define BRK_REGREADVAL   9   /* read specific value from register */
#define BRK_REGWRITEVAL  10  /* write specific value to register */
#ifdef USE_ADDRLIST
   #define BRK_ADDRLIST     11  /* Address not in valid list */
   #define NUM_BREAKTYPES   12
#else
   #define NUM_BREAKTYPES   11
#endif  /* USE_ADDRLIST */

#define NUM_WATCHREGS 24  /* all we have screen space for */
#define WATCH_DELAY   18  /* ticks per second, update watch once a second */

#define MAX_STIMULUS_SIZE 128   /* max size: (5 ports + DWORD) * 128 = 1152 bytes */
struct stim
   {
   DWORD cyclecount;
   WORD pinvals[NUM_BYTE_PORTS][8];
/*   BYTE pinvals[NUM_BYTE_PORTS]; */
   };

#define BROWSE_SCREEN_SIZE (24*80)

#ifdef PINS_8
   #define MAX_OSCOPE_LINES NUM_PORT_BITS
#else
   #define MAX_OSCOPE_LINES 8
#endif  /* PINS_8 */

struct picfilehdr
   {
   BYTE program_id[8];                                  /* 00 .. 07 */
   BYTE pictype[8];                                     /* 08 .. 0f */
   WORD filetype;   /* 0 = BIN, 1 = PIM */              /* 10 .. 11 */
   WORD version;    /* for both BIN and PIM */          /* 12 .. 13 */
         /* do not change anything above this point in later versions */
   WORD memsize;    /* for both BIN and PIM */          /* 14 .. 15 */
   WORD numregs;    /* PIM only */                      /* 16 .. 17 */
   WORD numeeprom;  /* PIM only */                      /* 18 .. 19 */
   WORD w_reg;      /* PIM only */                      /* 1a .. 1b */
   WORD ip_reg;     /* PIM only */                      /* 1c .. 1d */
   WORD stackptr;   /* PIM only */                      /* 1e .. 1f */
   WORD stackdepth; /* PIM only */                      /* 20 .. 21 */
   WORD maxstackd;  /* PIM only */                      /* 22 .. 23 */
   WORD stk[MAX_STACK_SIZE];  /* PIM only */            /* 24 .. 33 */
   BYTE tris;       /* PIM only, 12-bit */              /* 34 */
   BYTE option;     /* PIM only, 12-bit */              /* 35 */
   WORD config;     /* PIM only */                      /* 36 .. 37 */
   WORD scr_row;    /* PIM only */                      /* 38 .. 39 */
   WORD scr_col;    /* PIM only */                      /* 3a .. 3b */
   WORD filler[34]; /* make this 128 bytes long */      /* 3c .. 7f */
                    /* also future use */
   };
#ifdef PSHELL
   struct picfilehdr picfile_hdr;
#endif  /* PSHELL */
#define FILETYPE_BIN 0
#define FILETYPE_PIM 1
#define PICFILE_VERSION 2
   /* NOTE:  A BIN file contains the picfile_hdr structure, and the program
    *        memory (WORDS, stored lowbyte first).
    *
    *        A PIM (Pic IMage) file contains the picfile_hdr structure,
    *        the program memory (WORDS, stored lowbyte first), the register
    *        file, the EEPROM memory (optional, depends on processor), and
    *        the screen.
    */

#define CMD_NOTFOUND 0xffff  /* need this in both pshell.c and pshell2.c */

/*
 *  I2C stuff
 */
#define I2C_ONEBYTE_ADDR  2  /* control byte plus address byte */
#define I2C_OFF           0
#define I2C_WRITE         1
#define I2C_READ          2
#define I2C_BUFSIZE      68  /* max pagesize + 4 */
#define I2C_TYPE_128BIT   1  /* index into i2c_params[] for 24xx00 type */
struct processi2c
   {
   WORD hdr_size;   /* <control> + <1 or 2 address bytes> */
   WORD page_mask;  /* mask for page write */
   WORD addr_mask;  /* mask for WORD address */
   };

/* #define ANALOG_TRUE_LEVEL 0xcccc */  /* 0.8 Vdd, let's use Schmitt Trigger levels */
#define SCHMITT_TRUE_LEVEL 0xcccc  /* 0.8 Vdd */
#define SCHMITT_FALSE_LEVEL 0x3333  /* 0.2 Vdd */
#define CMOS_TRUE_LEVEL   0x6666    /* 0.4 Vdd for CMOS/TTL inputs */
/* NOTE: If desired later, each input pin could have it's own level
 *       (CMOS / Schmitt trigger / whatever).
 */
